import json
import mimetypes
import os
from string import Template
from typing import Sequence, Type

from fastapi.responses import Response
from langchain.schema.runnable import Runnable

from langserve.pydantic_v1 import BaseModel


class PlaygroundTemplate(Template):
    delimiter = "____"


def _get_mimetype(path: str) -> str:
    """Get mimetype for file.

    Custom implementation of mimetypes.guess_type that
    uses the file extension to determine the mimetype for some files.

    This is necessary due to: https://bugs.python.org/issue43975
    Resolves issue: https://github.com/langchain-ai/langserve/issues/245

    Args:
        path (str): Path to file

    Returns:
        str: Mimetype of file
    """
    try:
        file_extension = path.lower().split(".")[-1]
    except IndexError:
        return mimetypes.guess_type(path)[0]

    if file_extension == "js":
        return "application/javascript"
    elif file_extension == "css":
        return "text/css"
    elif file_extension in ["htm", "html"]:
        return "text/html"

    # If the file extension is not one of the specified ones,
    # use the default guess method
    mime_type = mimetypes.guess_type(path)[0]
    return mime_type


async def serve_playground(
    runnable: Runnable,
    input_schema: Type[BaseModel],
    config_keys: Sequence[str],
    base_url: str,
    file_path: str,
    feedback_enabled: bool,
) -> Response:
    """Serve the playground."""
    local_file_path = os.path.abspath(
        os.path.join(
            os.path.dirname(__file__),
            "./playground/dist",
            file_path or "index.html",
        )
    )

    base_dir = os.path.abspath(
        os.path.join(os.path.dirname(__file__), "./playground/dist")
    )

    if base_dir != os.path.commonpath((base_dir, local_file_path)):
        return Response("Not Found", status_code=404)

    try:
        with open(local_file_path, encoding="utf-8") as f:
            mime_type = _get_mimetype(local_file_path)
            if mime_type in ("text/html", "text/css", "application/javascript"):
                response = PlaygroundTemplate(f.read()).substitute(
                    LANGSERVE_BASE_URL=base_url[1:]
                    if base_url.startswith("/")
                    else base_url,
                    LANGSERVE_CONFIG_SCHEMA=json.dumps(
                        runnable.config_schema(include=config_keys).schema()
                    ),
                    LANGSERVE_INPUT_SCHEMA=json.dumps(input_schema.schema()),
                    LANGSERVE_FEEDBACK_ENABLED=json.dumps(
                        "true" if feedback_enabled else "false"
                    ),
                )
            else:
                response = f.buffer.read()
    except FileNotFoundError:
        return Response("Not Found", status_code=404)

    return Response(response, media_type=mime_type)



from datetime import datetime
from typing import Dict, List, Optional, Union
from uuid import UUID

from pydantic import BaseModel  # Floats between v1 and v2

from langserve.pydantic_v1 import BaseModel as BaseModelV1


class CustomUserType(BaseModelV1):
    """Inherit from this class to create a custom user type.

    Use a custom user type if you want the data to de-serialize
    into a pydantic model rather than the equivalent dict representation.

    In general, make sure to add a `type` attribute to your class
    to help pydantic to discriminate unions.

    https://docs.pydantic.dev/1.10/usage/types/#discriminated-unions-aka-tagged-unions

    Limitations:
        At the moment, this type only works SERVER side and is used
        to specify desired DECODING behavior. If inheriting from this type
        the server will keep the decoded type as a pydantic model instead
        of converting it into a dict.
    """


class SharedResponseMetadata(BaseModelV1):
    """
    Any response metadata should inherit from this class. Response metadata
    represents non-output data that may be useful to some clients, but
    ignorable to most. For example, the run_ids associated with each run
    kicked off by the associated request.

    SharedResponseMetadata is an abstraction to represent any metadata
    representing a LangServe response shared across all outputs in said
    response.
    """

    pass


class SingletonResponseMetadata(SharedResponseMetadata):
    """
    Represents response metadata used for just single input/output LangServe
    responses.
    """

    # Represents the parent run id for a given request
    run_id: UUID


class BatchResponseMetadata(SharedResponseMetadata):
    """
    Represents response metadata used for batches of input/output LangServe
    responses.
    """

    # Represents each parent run id for a given request, in
    # the same order in which they were received
    run_ids: List[UUID]


class BaseFeedback(BaseModel):
    """
    Shared information between create requests of feedback and feedback objects
    """

    run_id: UUID
    """The associated run ID this feedback is logged for."""

    key: str
    """The metric name, tag, or aspect to provide feedback on."""

    score: Optional[Union[float, int, bool]] = None
    """Value or score to assign the run."""

    value: Optional[Union[float, int, bool, str, Dict]] = None
    """The display value for the feedback if not a metric."""

    comment: Optional[str] = None
    """Comment or explanation for the feedback."""


class FeedbackCreateRequest(BaseFeedback):
    """
    Represents a request that creates feedback for an individual run
    """

    pass


class Feedback(BaseFeedback):
    """
    Represents feedback given on an individual run
    """

    id: UUID
    """The unique ID of the feedback that was created."""

    created_at: datetime
    """The time the feedback was created."""

    modified_at: datetime
    """The time the feedback was last modified."""

    correction: Optional[Dict] = None
    """Correction for the run."""



"""FastAPI integration for langchain runnables.

This code contains integration for langchain runnables with FastAPI.

The main entry point is the `add_routes` function which adds the routes to an existing
FastAPI app or APIRouter.
"""
import weakref
from typing import (
    Any,
    Literal,
    Optional,
    Sequence,
    Type,
    Union,
)

from langchain.schema.runnable import Runnable
from typing_extensions import Annotated

from langserve.api_handler import PerRequestConfigModifier, _APIHandler, _is_hosted
from langserve.pydantic_v1 import (
    _PYDANTIC_MAJOR_VERSION,
    PYDANTIC_VERSION,
    BaseModel,
)

try:
    from fastapi import APIRouter, FastAPI
except ImportError:
    # [server] extra not installed
    APIRouter = FastAPI = Any

# A function that that takes a config and a raw request
# and updates the config based on the request.

# This is a global registry of models to avoid creating the same model
# multiple times.
# Duplicated model names break fastapi's openapi generation.

_APP_SEEN = weakref.WeakSet()
_APP_TO_PATHS = weakref.WeakKeyDictionary()


def _setup_global_app_handlers(app: Union[FastAPI, APIRouter]) -> None:
    @app.on_event("startup")
    async def startup_event():
        GIGASERVE = r"""
  _______  __    _______      ___           _______. _______ .______     ____    ____  _______ 
 /  _____||  |  /  _____|    /   \         /       ||   ____||   _  \    \   \  /   / |   ____|
|  |  __  |  | |  |  __     /  ^  \       |   (----`|  |__   |  |_)  |    \   \/   /  |  |__   
|  | |_ | |  | |  | |_ |   /  /_\  \       \   \    |   __|  |      /      \      /   |   __|  
|  |__| | |  | |  |__| |  /  _____  \  .----)   |   |  |____ |  |\  \----.  \    /    |  |____ 
 \______| |__|  \______| /__/     \__\ |_______/    |_______|| _| `._____|   \__/     |_______|
"""  # noqa: E501

        def green(text: str) -> str:
            """Return the given text in green."""
            return "\x1b[1;32;40m" + text + "\x1b[0m"

        def orange(text: str) -> str:
            """Return the given text in orange."""
            return "\x1b[1;31;40m" + text + "\x1b[0m"

        paths = _APP_TO_PATHS[app]
        print(GIGASERVE)
        for path in paths:
            print(
                f'{green("GIGASERVE:")} Playground for chain "{path or ""}/" is '
                f'live at:'
            )
            print(f'{green("GIGASERVE:")}  │')
            print(f'{green("GIGASERVE:")}  └──> {path}/playground/')
            print(f'{green("GIGASERVE:")}')
        print(f'{green("GIGASERVE:")} See all available routes at {app.docs_url}/')

        if _PYDANTIC_MAJOR_VERSION == 2:
            print()
            print(f'{orange("GIGASERVE:")} ', end="")
            print(
                f"⚠️ Using pydantic {PYDANTIC_VERSION}. "
                f"OpenAPI docs for invoke, batch, stream, stream_log "
                f"endpoints will not be generated. API endpoints and playground "
                f"should work as expected. "
                f"If you need to see the docs, you can downgrade to pydantic 1. "
                "For example, `pip install pydantic==1.10.13`. "
                f"See https://github.com/tiangolo/fastapi/issues/10360 for details."
            )
        print()


def _register_path_for_app(app: Union[FastAPI, APIRouter], path: str) -> None:
    """Register a path when its added to app. Raise if path already seen."""
    if app in _APP_TO_PATHS:
        seen_paths = _APP_TO_PATHS.get(app)
        if path in seen_paths:
            raise ValueError(
                f"A runnable already exists at path: {path}. If adding "
                f"multiple runnables make sure they have different paths."
            )
        seen_paths.add(path)
    else:
        _setup_global_app_handlers(app)
        _APP_TO_PATHS[app] = {path}


EndpointName = Literal[
    "invoke",
    "batch",
    "stream",
    "stream_log",
    "playground",
    "feedback",
    "input_schema",
    "config_schema",
    "output_schema",
    "config_hashes",
]

KNOWN_ENDPOINTS = {
    "invoke",
    "batch",
    "stream",
    "stream_log",
    "playground",
    "feedback",
    "input_schema",
    "config_schema",
    "output_schema",
    "config_hashes",
}


class _EndpointConfiguration:
    """Logic for enabling/disabling endpoints."""

    def __init__(
        self,
        *,
        enabled_endpoints: Optional[Sequence[EndpointName]] = None,
        disabled_endpoints: Optional[Sequence[EndpointName]] = None,
        enable_feedback_endpoint: bool = False,
    ) -> None:
        """Initialize the endpoint configuration."""
        if enabled_endpoints and disabled_endpoints:
            raise ValueError(
                f'Cannot specify both "enabled_endpoints" and "disabled_endpoints".'
                f"Got enabled_endpoints={enabled_endpoints} and disabled_endpoints="
                f"{disabled_endpoints}."
            )

        if enabled_endpoints and not isinstance(enabled_endpoints, Sequence):
            raise ValueError(
                f"Expected enabled_endpoints to be a sequence (e.g., list or tuple), "
                f"got {type(enabled_endpoints)}"
            )

        if disabled_endpoints and not isinstance(disabled_endpoints, Sequence):
            raise ValueError(
                f"Expected disabled_endpoints to be a sequence (e.g., list or tuple), "
                f"got {type(disabled_endpoints)}"
            )

        if enabled_endpoints is None:
            if disabled_endpoints is None:
                is_invoke_enabled = True
                is_batch_enabled = True
                is_stream_enabled = True
                is_stream_log_enabled = True
                is_playground_enabled = True
                is_input_schema_enabled = True
                is_output_schema_enabled = True
                is_config_schema_enabled = True
                is_config_hash_enabled = True
            else:
                disabled_endpoints_ = set(name.lower() for name in disabled_endpoints)
                if disabled_endpoints_ - KNOWN_ENDPOINTS:
                    raise ValueError(
                        f"Got unknown endpoint "
                        f"names: {disabled_endpoints_ - KNOWN_ENDPOINTS}"
                    )
                is_invoke_enabled = "invoke" not in disabled_endpoints_
                is_batch_enabled = "batch" not in disabled_endpoints_
                is_stream_enabled = "stream" not in disabled_endpoints_
                is_stream_log_enabled = "stream_log" not in disabled_endpoints_
                is_playground_enabled = "playground" not in disabled_endpoints_
                is_input_schema_enabled = "input_schema" not in disabled_endpoints_
                is_output_schema_enabled = "output_schema" not in disabled_endpoints_
                is_config_schema_enabled = "config_schema" not in disabled_endpoints_
                is_config_hash_enabled = "config_hashes" not in disabled_endpoints_
        else:
            enabled_endpoints_ = set(name.lower() for name in enabled_endpoints)
            if enabled_endpoints_ - KNOWN_ENDPOINTS:
                raise ValueError(
                    f"Got unknown endpoint names: {enabled_endpoints_- KNOWN_ENDPOINTS}"
                )
            is_invoke_enabled = "invoke" in enabled_endpoints_
            is_batch_enabled = "batch" in enabled_endpoints_
            is_stream_enabled = "stream" in enabled_endpoints_
            is_stream_log_enabled = "stream_log" in enabled_endpoints_
            is_playground_enabled = "playground" in enabled_endpoints_
            is_input_schema_enabled = "input_schema" in enabled_endpoints_
            is_output_schema_enabled = "output_schema" in enabled_endpoints_
            is_config_schema_enabled = "config_schema" in enabled_endpoints_
            is_config_hash_enabled = "config_hashes" in enabled_endpoints_

        self.is_invoke_enabled = is_invoke_enabled
        self.is_batch_enabled = is_batch_enabled
        self.is_stream_enabled = is_stream_enabled
        self.is_stream_log_enabled = is_stream_log_enabled
        self.is_playground_enabled = is_playground_enabled
        self.is_input_schema_enabled = is_input_schema_enabled
        self.is_output_schema_enabled = is_output_schema_enabled
        self.is_config_type_enabled = is_config_schema_enabled
        self.is_config_hash_enabled = is_config_hash_enabled
        self.is_feedback_enabled = enable_feedback_endpoint


# PUBLIC API


def add_routes(
    app: Union[FastAPI, APIRouter],
    runnable: Runnable,
    *,
    path: str = "",
    input_type: Union[Type, Literal["auto"], BaseModel] = "auto",
    output_type: Union[Type, Literal["auto"], BaseModel] = "auto",
    config_keys: Sequence[str] = ("configurable",),
    include_callback_events: bool = False,
    per_req_config_modifier: Optional[PerRequestConfigModifier] = None,
    enable_feedback_endpoint: bool = _is_hosted(),
    disabled_endpoints: Optional[Sequence[EndpointName]] = None,
    stream_log_name_allow_list: Optional[Sequence[str]] = None,
    enabled_endpoints: Optional[Sequence[EndpointName]] = None,
) -> None:
    """Register the routes on the given FastAPI app or APIRouter.


    The following routes are added per runnable under the specified `path`:

    * /invoke - for invoking a runnable with a single input
    * /batch - for invoking a runnable with multiple inputs
    * /stream - for streaming the output of a runnable
    * /stream_log - for streaming intermediate outputs for a runnable
    * /input_schema - for returning the input schema of the runnable
    * /output_schema - for returning the output schema of the runnable
    * /config_schema - for returning the config schema of the runnable

    Args:
        app: The FastAPI app or APIRouter to which routes should be added.
        runnable: The runnable to wrap, must not be stateful.
        path: A path to prepend to all routes.
        input_type: type to use for input validation.
            Default is "auto" which will use the InputType of the runnable.
            User is free to provide a custom type annotation.
            Favor using runnable.with_types(input_type=..., output_type=...) instead.
            This parameter may get deprecated!
        output_type: type to use for output validation.
            Default is "auto" which will use the OutputType of the runnable.
            User is free to provide a custom type annotation.
            Favor using runnable.with_types(input_type=..., output_type=...) instead.
            This parameter may get deprecated!
        config_keys: list of config keys that will be accepted, by default
            will accept `configurable` key in the config. Will only be used
            if the runnable is configurable. Cannot configure run_name,
            which is set by default to the path of the API.
        include_callback_events: Whether to include callback events in the response.
            If true, the client will be able to show trace information
            including events that occurred on the server side.
            Be sure not to include any sensitive information in the callback events.
        per_req_config_modifier: optional function that can be used to update the
            RunnableConfig for a given run based on the raw request. This is useful,
            for example, if the user wants to pass in a header containing credentials
            to a runnable. The RunnableConfig is presented in its dictionary form.
            Note that only keys in `config_keys` will be modifiable by this function.
        enable_feedback_endpoint: Whether to enable an endpoint for logging feedback
            to LangSmith. Enabled by default. If this flag is disabled or LangSmith
            tracing is not enabled for the runnable, then 400 errors will be thrown
            when accessing the feedback endpoint
        enabled_endpoints: A list of endpoints which should be enabled. If not
            specified, all associated endpoints will be enabled. The list can contain
            the following values: *invoke*, *batch*, *stream*, *stream_log*,
            *playground*, *input_schema*, *output_schema*, *config_schema*,
            *config_hashes*.

            *config_hashes* represents the config hash variant (when it exists)
            of each endpoint. Enabling this is useful when working with configurable
            runnables and sharing playground configuration links to the runnables.

            For example, if we want to enable regular invoke and batch endpoints
            and their config hash variants, we can do:

            ```python

            add_routes(
                ...,
                enabled_endpoints=("invoke", "batch", "config_hashes"),
            )
            ```

            Please note that the feedback endpoint is not included in this list
            and is controlled by the `enable_feedback_endpoint` flag.
        disabled_endpoints: A list of endpoints which should be disabled. If not
            specified, all associated endpoints will be enabled. The list can contain
            the following values: *invoke*, *batch*, *stream*, *stream_log*,
            *playground*, *input_schema*, *output_schema*, *config_schema*,
            *config_hashes*.

            *config_hashes* represents the config hash variant (when it exists)
            of each endpoint. Enabling this is useful when working with configurable
            runnables and sharing playground configuration links to the runnables.

            For example, if we want to enable regular invoke and batch endpoints
            and their config hash variants, we can do:

            ```python

            add_routes(
                ...,
                disabled_endpoints=["playground"],
            )
            ```
        stream_log_name_allow_list: list of run names that the client can
            stream as intermediate steps

    """
    endpoint_configuration = _EndpointConfiguration(
        enabled_endpoints=enabled_endpoints,
        disabled_endpoints=disabled_endpoints,
        enable_feedback_endpoint=enable_feedback_endpoint,
    )

    try:
        from sse_starlette import EventSourceResponse
    except ImportError:
        raise ImportError(
            "sse_starlette must be installed to implement the stream and "
            "stream_log endpoints. "
            "Use `pip install sse_starlette` to install."
        )

    if path and not path.startswith("/"):
        raise ValueError(
            f"Got an invalid path: {path}. "
            f"If specifying path please start it with a `/`"
        )

    if isinstance(app, FastAPI):  # type: ignore
        # Cannot do this checking logic for a router since
        # API routers are not hashable
        _register_path_for_app(app, path)

    # Determine the base URL for the playground endpoint
    prefix = app.prefix if isinstance(app, APIRouter) else ""  # type: ignore

    api_handler = _APIHandler(
        runnable,
        path=path,
        prefix=prefix,
        input_type=input_type,
        output_type=output_type,
        config_keys=config_keys,
        include_callback_events=include_callback_events,
        enable_feedback_endpoint=enable_feedback_endpoint,
        per_req_config_modifier=per_req_config_modifier,
        stream_log_name_allow_list=stream_log_name_allow_list,
    )
    namespace = path or ""

    route_tags = [path.strip("/")] if path else None
    route_tags_with_config = [f"{path.strip('/')}/config"] if path else ["config"]

    def _route_name(name: str) -> str:
        """Return the route name with the given name."""
        return f"{path.strip('/')} {name}" if path else name

    def _route_name_with_config(name: str) -> str:
        """Return the route name with the given name."""
        return (
            f"{path.strip('/')} {name} with config" if path else f"{name} with config"
        )

    if hasattr(app, "openapi_tags") and (path or (app not in _APP_SEEN)):
        if not path:
            _APP_SEEN.add(app)

        if _PYDANTIC_MAJOR_VERSION == 1:
            # Documentation for the default endpoints
            default_endpoint_tags = {
                "name": route_tags[0] if route_tags else "default",
            }
        elif _PYDANTIC_MAJOR_VERSION == 2:
            # When using pydantic v2, we cannot generate openapi docs for
            # the invoke/batch/stream/stream_log endpoints since the underlying
            # models are from the pydantic.v1 namespace and cannot be supported
            # by fastapi's.
            # https://github.com/tiangolo/fastapi/issues/10360
            default_endpoint_tags = {
                "name": route_tags[0] if route_tags else "default",
                "description": (
                    f"⚠️ Using pydantic {PYDANTIC_VERSION}. "
                    f"OpenAPI docs for `invoke`, `batch`, `stream`, `stream_log` "
                    f"endpoints will not be generated. API endpoints and playground "
                    f"should work as expected. "
                    f"If you need to see the docs, you can downgrade to pydantic 1. "
                    "For example, `pip install pydantic==1.10.13`"
                    f"See https://github.com/tiangolo/fastapi/issues/10360 for details."
                ),
            }
        else:
            raise AssertionError(
                f"Expected pydantic major version 1 or 2, got {_PYDANTIC_MAJOR_VERSION}"
            )

        if endpoint_configuration.is_config_hash_enabled:
            app.openapi_tags = [
                *(getattr(app, "openapi_tags", []) or []),
                default_endpoint_tags,
                {
                    "name": route_tags_with_config[0],
                    "description": (
                        "Endpoints with a default configuration "
                        "set by `config_hash` path parameter. "
                        "Used in conjunction with share links generated using the "
                        "LangServe UI playground. "
                        "The hash is an LZString compressed JSON string."
                    ),
                },
            ]

    if endpoint_configuration.is_invoke_enabled:
        invoke = app.post(f"{namespace}/invoke", include_in_schema=False)(
            api_handler.invoke
        )

        if endpoint_configuration.is_config_hash_enabled:
            app.post(
                namespace + "/c/{config_hash}/invoke",
                include_in_schema=False,
            )(invoke)

    if endpoint_configuration.is_batch_enabled:
        batch = app.post(f"{namespace}/batch", include_in_schema=False)(
            api_handler.batch
        )

        if endpoint_configuration.is_config_hash_enabled:
            app.post(
                namespace + "/c/{config_hash}/batch",
                include_in_schema=False,
            )(batch)

    if endpoint_configuration.is_stream_enabled:
        stream = app.post(f"{namespace}/stream", include_in_schema=False)(
            api_handler.stream
        )

        if endpoint_configuration.is_config_hash_enabled:
            app.post(
                namespace + "/c/{config_hash}/stream",
                include_in_schema=False,
            )(stream)

    if endpoint_configuration.is_stream_log_enabled:
        stream_log = app.post(f"{namespace}/stream_log", include_in_schema=False)(
            api_handler.stream_log
        )

        if endpoint_configuration.is_config_hash_enabled:
            app.post(
                namespace + "/c/{config_hash}/stream_log",
                include_in_schema=False,
            )(stream_log)

    if endpoint_configuration.is_input_schema_enabled:
        input_schema = app.get(
            f"{namespace}/input_schema",
            tags=route_tags,
            name=_route_name("input_schema"),
        )(api_handler.input_schema)

        if endpoint_configuration.is_config_hash_enabled:
            app.get(
                namespace + "/c/{config_hash}/input_schema",
                tags=route_tags_with_config,
                name=_route_name_with_config("input_schema"),
            )(input_schema)

    if endpoint_configuration.is_output_schema_enabled:
        output_schema = app.get(
            f"{namespace}/output_schema",
            tags=route_tags,
            name=_route_name("output_schema"),
        )(api_handler.output_schema)

        if endpoint_configuration.is_config_hash_enabled:
            app.get(
                namespace + "/c/{config_hash}/output_schema",
                tags=route_tags_with_config,
                name=_route_name_with_config("output_schema"),
            )(output_schema)

    if endpoint_configuration.is_input_schema_enabled:
        config_schema = app.get(
            f"{namespace}/config_schema",
            tags=route_tags,
            name=_route_name("config_schema"),
        )(api_handler.config_schema)

        if endpoint_configuration.is_config_hash_enabled:
            app.get(
                namespace + "/c/{config_hash}/config_schema",
                tags=route_tags_with_config,
                name=_route_name_with_config("config_schema"),
            )(config_schema)

    if endpoint_configuration.is_playground_enabled:
        playground = app.get(namespace + "/playground/{file_path:path}")(
            api_handler.playground
        )

        if endpoint_configuration.is_config_hash_enabled:
            app.get(
                namespace + "/c/{config_hash}/playground/{file_path:path}",
            )(playground)

    if enable_feedback_endpoint:
        create_feedback = app.post(
            namespace + "/feedback",
        )(api_handler.create_feedback)

        if endpoint_configuration.is_config_hash_enabled:
            app.post(
                namespace + "/c/{config_hash}/feedback",
            )(create_feedback)

        check_feedback_enabled = app.head(
            namespace + "/feedback",
        )(api_handler._check_feedback_enabled)

        # Add get version which returns True/False for flow control
        # instead of 200/400 for flow control
        app.get(
            namespace + "/feedback",
        )(api_handler.check_feedback_enabled)

        if endpoint_configuration.is_config_hash_enabled:  # Is this needed?
            app.head(
                namespace + "/c/{config_hash}/feedback",
            )(check_feedback_enabled)

    #######################################
    # Documentation variants of end points.
    #######################################
    # At the moment, we only support pydantic 1.x for documentation
    if _PYDANTIC_MAJOR_VERSION == 1:
        InvokeRequest = api_handler.InvokeRequest
        InvokeResponse = api_handler.InvokeResponse
        BatchRequest = api_handler.BatchRequest
        BatchResponse = api_handler.BatchResponse
        StreamRequest = api_handler.StreamRequest
        StreamLogRequest = api_handler.StreamLogRequest

        if endpoint_configuration.is_invoke_enabled:

            async def _invoke_docs(
                invoke_request: Annotated[InvokeRequest, InvokeRequest],
                config_hash: str = "",
            ) -> InvokeResponse:
                """Invoke the runnable with the given input and config."""
                raise AssertionError("This endpoint should not be reachable.")

            invoke_docs = app.post(
                f"{namespace}/invoke",
                response_model=api_handler.InvokeResponse,
                tags=route_tags,
                name=_route_name("invoke"),
            )(_invoke_docs)

            if endpoint_configuration.is_config_hash_enabled:
                app.post(
                    namespace + "/c/{config_hash}/invoke",
                    response_model=api_handler.InvokeResponse,
                    tags=route_tags_with_config,
                    name=_route_name_with_config("invoke"),
                    description=(
                        "This endpoint is to be used with share links generated by the "
                        "LangServe playground. "
                        "The hash is an LZString compressed JSON string. "
                        "For regular use cases, use the /invoke endpoint without "
                        "the `c/{config_hash}` path parameter."
                    ),
                )(invoke_docs)

        if endpoint_configuration.is_batch_enabled:

            async def _batch_docs(
                batch_request: Annotated[BatchRequest, BatchRequest],
                config_hash: str = "",
            ) -> BatchResponse:
                """Batch invoke the runnable with the given inputs and config."""
                raise AssertionError("This endpoint should not be reachable.")

            batch_docs = app.post(
                f"{namespace}/batch",
                response_model=BatchResponse,
                tags=route_tags,
                name=_route_name("batch"),
            )(_batch_docs)

            if endpoint_configuration.is_config_hash_enabled:
                app.post(
                    namespace + "/c/{config_hash}/batch",
                    response_model=BatchResponse,
                    tags=route_tags_with_config,
                    name=_route_name_with_config("batch"),
                    description=(
                        "This endpoint is to be used with share links generated by the "
                        "LangServe playground. "
                        "The hash is an LZString compressed JSON string. "
                        "For regular use cases, use the /batch endpoint without "
                        "the `c/{config_hash}` path parameter."
                    ),
                )(batch_docs)

        if endpoint_configuration.is_stream_enabled:

            async def _stream_docs(
                stream_request: Annotated[StreamRequest, StreamRequest],
                config_hash: str = "",
            ) -> EventSourceResponse:
                """Invoke the runnable stream the output.

                This endpoint allows to stream the output of the runnable.

                The endpoint uses a server sent event stream to stream the output.

                https://developer.mozilla.org/en-US/docs/Web/API/Server-sent_events

                Important: Set the "text/event-stream" media type for request headers if
                           not using an existing SDK.

                This endpoint uses two different types of events:

                * data - for streaming the output of the runnable

                    {
                        "event": "data",
                        "data": {
                        ...
                        }
                    }

                * error - for signaling an error in the stream, also ends the stream.

                {
                    "event": "error",
                    "data": {
                        "status_code": 500,
                        "message": "Internal Server Error"
                    }
                }

                * end - for signaling the end of the stream.

                    This helps the client to know when to stop listening for events and
                    know that the streaming has ended successfully.

                    {
                        "event": "end",
                    }
                """
                raise AssertionError("This endpoint should not be reachable.")

            stream_docs = app.post(
                f"{namespace}/stream",
                include_in_schema=True,
                tags=route_tags,
                name=_route_name("stream"),
            )(_stream_docs)

            if endpoint_configuration.is_config_hash_enabled:
                app.post(
                    namespace + "/c/{config_hash}/stream",
                    include_in_schema=True,
                    tags=route_tags_with_config,
                    name=_route_name_with_config("stream"),
                    description=(
                        "This endpoint is to be used with share links generated by the "
                        "LangServe playground. "
                        "The hash is an LZString compressed JSON string. "
                        "For regular use cases, use the /stream endpoint without "
                        "the `c/{config_hash}` path parameter."
                    ),
                )(stream_docs)

        if endpoint_configuration.is_stream_log_enabled:

            async def _stream_log_docs(
                stream_log_request: Annotated[StreamLogRequest, StreamLogRequest],
                config_hash: str = "",
            ) -> EventSourceResponse:
                """Invoke the runnable stream_log the output.

                This endpoint allows to stream the output of the runnable, including
                the output of all intermediate steps.

                The endpoint uses a server sent event stream to stream the output.

                https://developer.mozilla.org/en-US/docs/Web/API/Server-sent_events

                Important: Set the "text/event-stream" media type for request headers if
                           not using an existing SDK.

                This endpoint uses two different types of events:

                * data - for streaming the output of the runnable

                    {
                        "event": "data",
                        "data": {
                        ...
                        }
                    }

                * error - for signaling an error in the stream, also ends the stream.

                {
                    "event": "error",
                    "data": {
                        "status_code": 500,
                        "message": "Internal Server Error"
                    }
                }

                * end - for signaling the end of the stream.

                    This helps the client to know when to stop listening for events and
                    know that the streaming has ended successfully.

                    {
                        "event": "end",
                    }
                """
                raise AssertionError("This endpoint should not be reachable.")

            app.post(
                f"{namespace}/stream_log",
                include_in_schema=True,
                tags=route_tags,
                name=_route_name("stream_log"),
            )(_stream_log_docs)

            if endpoint_configuration.is_config_hash_enabled:
                app.post(
                    namespace + "/c/{config_hash}/stream_log",
                    include_in_schema=True,
                    tags=route_tags_with_config,
                    name=_route_name_with_config("stream_log"),
                    description=(
                        "This endpoint is to be used with share links generated by the "
                        "LangServe playground. "
                        "The hash is an LZString compressed JSON string. "
                        "For regular use cases, use the /stream_log endpoint without "
                        "the `c/{config_hash}` path parameter."
                    ),
                )(_stream_log_docs)



import contextlib
import importlib
import json
import os
import re
from inspect import isclass
from typing import (
    Any,
    AsyncIterator,
    Callable,
    Dict,
    Generator,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

from fastapi import HTTPException
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import RequestValidationError
from langchain.callbacks.base import AsyncCallbackHandler
from langchain.callbacks.tracers.log_stream import RunLogPatch
from langchain.load.serializable import Serializable
from langchain.schema.runnable import Runnable, RunnableConfig
from langchain.schema.runnable.config import get_config_list, merge_configs
from langsmith import client as ls_client
from langsmith.utils import tracing_is_enabled
from starlette.requests import Request
from starlette.responses import JSONResponse, Response

from langserve.callbacks import AsyncEventAggregatorCallback, CallbackEventDict
from langserve.lzstring import LZString
from langserve.playground import serve_playground
from langserve.pydantic_v1 import BaseModel, Field, ValidationError, create_model
from langserve.schema import (
    BatchResponseMetadata,
    CustomUserType,
    Feedback,
    FeedbackCreateRequest,
    SingletonResponseMetadata,
)
from langserve.serialization import WellKnownLCSerializer
from langserve.validation import (
    BatchBaseResponse,
    BatchRequestShallowValidator,
    InvokeBaseResponse,
    InvokeRequestShallowValidator,
    StreamLogParameters,
    create_batch_request_model,
    create_batch_response_model,
    create_invoke_request_model,
    create_invoke_response_model,
    create_stream_log_request_model,
    create_stream_request_model,
)
from langserve.version import __version__

try:
    from sse_starlette import EventSourceResponse
except ImportError:
    EventSourceResponse = Any


def _is_hosted() -> bool:
    return os.environ.get("HOSTED_LANGSERVE_ENABLED", "false").lower() == "true"


def _config_from_hash(config_hash: str) -> Dict[str, Any]:
    try:
        if not config_hash:
            return {}

        uncompressed = LZString.decompressFromEncodedURIComponent(config_hash)
        parsed = json.loads(uncompressed)
        if isinstance(parsed, dict):
            return parsed
        else:
            raise HTTPException(400, "Invalid config hash")
    except Exception:
        raise HTTPException(400, "Invalid config hash")


PerRequestConfigModifier = Callable[[Dict[str, Any], Request], Dict[str, Any]]


def _unpack_request_config(
    *configs: Union[BaseModel, Mapping, str],
    config_keys: Sequence[str],
    model: Type[BaseModel],
    request: Request,
    per_req_config_modifier: Optional[PerRequestConfigModifier],
) -> Dict[str, Any]:
    """Merge configs, and project the given keys from the merged dict."""
    config_dicts = []
    for config in configs:
        if isinstance(config, str):
            config_dicts.append(model(**_config_from_hash(config)).dict())
        elif isinstance(config, BaseModel):
            config_dicts.append(config.dict())
        elif isinstance(config, Mapping):
            config_dicts.append(model(**config).dict())
        else:
            raise TypeError(f"Expected a string, dict or BaseModel got {type(config)}")
    config = merge_configs(*config_dicts)
    if "configurable" in config and config["configurable"]:
        if "configurable" not in config_keys:
            raise HTTPException(
                422,
                "The config field `configurable` has been disallowed by the server. "
                "This can be modified server side by adding `configurable` to the list "
                "of `config_keys` argument in `add_routes`",
            )
    projected_config = {k: config[k] for k in config_keys if k in config}
    return (
        per_req_config_modifier(projected_config, request)
        if per_req_config_modifier
        else projected_config
    )


def _update_config_with_defaults(
    path: str,
    incoming_config: RunnableConfig,
    request: Request,
    *,
    endpoint: Optional[str] = None,
) -> RunnableConfig:
    """Set up some baseline configuration for the underlying runnable."""

    # Currently all defaults are non-overridable
    overridable_default_config = RunnableConfig()

    metadata = {
        "__useragent": request.headers.get("user-agent"),
        "__langserve_version": __version__,
    }

    if endpoint:
        metadata["__langserve_endpoint"] = endpoint

    if _is_hosted():
        hosted_metadata = {
            "__langserve_hosted_git_commit_sha": os.environ.get(
                "HOSTED_LANGSERVE_GIT_COMMIT", ""
            ),
            "__langserve_hosted_repo_subdirectory_path": os.environ.get(
                "HOSTED_LANGSERVE_GIT_REPO_PATH", ""
            ),
            "__langserve_hosted_repo_url": os.environ.get(
                "HOSTED_LANGSERVE_GIT_REPO", ""
            ),
            "__langserve_hosted_is_hosted": "true",
        }
        metadata.update(hosted_metadata)

    non_overridable_default_config = RunnableConfig(
        run_name=path,
        metadata=metadata,
    )

    # merge_configs is last-writer-wins, so we specifically pass in the
    # overridable configs first, then the user provided configs, then
    # finally the non-overridable configs
    return merge_configs(
        overridable_default_config,
        incoming_config,
        non_overridable_default_config,
    )


def _unpack_input(validated_model: BaseModel) -> Any:
    """Unpack the decoded input from the validated model."""
    if hasattr(validated_model, "__root__"):
        model = validated_model.__root__
    else:
        model = validated_model

    if isinstance(model, BaseModel) and not isinstance(
        model, (Serializable, CustomUserType)
    ):
        # If the model is a pydantic model, but not a Serializable, then
        # it was created by the server as part of validation and isn't expected
        # to be accepted by the runnables as input as a pydantic model,
        # instead we need to convert it into a corresponding python dict.
        # This logic should be applied recursively to nested models.
        return {
            fieldname: _unpack_input(getattr(model, fieldname))
            for fieldname in model.__fields__.keys()
        }

    return model


def _rename_pydantic_model(model: Type[BaseModel], prefix: str) -> Type[BaseModel]:
    """Rename the given pydantic model to the given name."""
    return create_model(
        prefix + model.__name__,
        __config__=model.__config__,
        **{
            fieldname: (
                _rename_pydantic_model(field.annotation, prefix)
                if isclass(field.annotation) and issubclass(field.annotation, BaseModel)
                else field.annotation,
                Field(
                    field.default,
                    title=fieldname,
                    description=field.field_info.description,
                ),
            )
            for fieldname, field in model.__fields__.items()
        },
    )


def _replace_non_alphanumeric_with_underscores(s: str) -> str:
    """Replace non-alphanumeric characters with underscores."""
    return re.sub(r"[^a-zA-Z0-9]", "_", s)


def _resolve_model(
    type_: Union[Type, BaseModel], default_name: str, namespace: str
) -> Type[BaseModel]:
    """Resolve the input type to a BaseModel."""
    if isclass(type_) and issubclass(type_, BaseModel):
        model = type_
    else:
        model = create_model(default_name, __root__=(type_, ...))

    hash_ = model.schema_json()

    if model.__name__ in _SEEN_NAMES and hash_ not in _MODEL_REGISTRY:
        # If the model name has been seen before, but the model itself is different
        # generate a new name for the model.
        model_to_use = _rename_pydantic_model(model, namespace)
        hash_ = model_to_use.schema_json()
    else:
        model_to_use = model

    if hash_ not in _MODEL_REGISTRY:
        _SEEN_NAMES.add(model_to_use.__name__)
        _MODEL_REGISTRY[hash_] = model_to_use

    return _MODEL_REGISTRY[hash_]


def _add_namespace_to_model(namespace: str, model: Type[BaseModel]) -> Type[BaseModel]:
    """Prefix the name of the given model with the given namespace.

    Code is used to help avoid name collisions when hosting multiple runnables
    that may use the same underlying models.

    Args:
        namespace: The namespace to use for the model.
        model: The model to create a unique name for.

    Returns:
        A new model with name prepended with the given namespace.
    """
    model_with_unique_name = _rename_pydantic_model(model, namespace)
    model_with_unique_name.update_forward_refs()
    return model_with_unique_name


def _scrub_exceptions_in_event(event: CallbackEventDict) -> CallbackEventDict:
    """Scrub exceptions and change to a serializable format."""
    type_ = event["type"]
    # Check if the event type is one that could contain an error key
    # for example, on_chain_error, on_tool_error, etc.
    if "error" not in type_:
        return event

    # This is not scrubbing -- it's doing serialization
    if "error" not in event:  # if there is an error key, scrub it
        return event

    if isinstance(event["error"], BaseException):
        event.copy()
        event["error"] = {"status_code": 500, "message": "Internal Server Error"}
        return event

    raise AssertionError(f"Expected an exception got {type(event['error'])}")


@contextlib.contextmanager
def _with_validation_error_translation() -> Generator[None, None, None]:
    """Context manager to translate validation errors to request validation errors.

    This makes sure that validation errors are surfaced as client side errors.
    """
    try:
        yield
    except ValidationError as e:
        raise RequestValidationError(e.errors(), body=e.model)


def _get_base_run_id_as_str(
    event_aggregator: AsyncEventAggregatorCallback,
) -> Optional[str]:
    """
    Uses `event_aggregator` to determine the base run ID for a given run. Returns
    the run_id as a string, or None if it does not exist.
    """
    # The first run in the callback_events list corresponds to the
    # overall trace for request
    if event_aggregator.callback_events and event_aggregator.callback_events[0].get(
        "run_id"
    ):
        return str(event_aggregator.callback_events[0].get("run_id"))
    else:
        raise AssertionError("No run_id found for the given run")


def _json_encode_response(model: BaseModel) -> JSONResponse:
    """Return a JSONResponse with the given content.

    We're doing the encoding manually here as a workaround to fastapi
    not supporting models from pydantic v1 when pydantic
    v2 is imported.

    Args:
        obj: The object to encode; either an invoke response or a batch response.

    Returns:
        A JSONResponse with the given content.
    """
    obj = jsonable_encoder(model)

    if isinstance(model, InvokeBaseResponse):
        # Invoke Response
        # Collapse '__root__' from output field if it exists. This is done
        # automatically by fastapi when annotating request and response with
        # We need to do this manually since we're using vanilla JSONResponse
        if isinstance(obj["output"], dict) and "__root__" in obj["output"]:
            obj["output"] = obj["output"]["__root__"]

        if "callback_events" in obj:
            for idx, callback_event in enumerate(obj["callback_events"]):
                if isinstance(callback_event, dict) and "__root__" in callback_event:
                    obj["callback_events"][idx] = callback_event["__root__"]
    elif isinstance(model, BatchBaseResponse):
        if not isinstance(obj["output"], list):
            raise AssertionError("Expected output to be a list")

        # Collapse '__root__' from output field if it exists. This is done
        # automatically by fastapi when annotating request and response with
        # We need to do this manually since we're using vanilla JSONResponse
        outputs = obj["output"]
        for idx, output in enumerate(outputs):
            if isinstance(output, dict) and "__root__" in output:
                outputs[idx] = output["__root__"]

        if "callback_events" in obj:
            if not isinstance(obj["callback_events"], list):
                raise AssertionError("Expected callback_events to be a list")

            for callback_events in obj["callback_events"]:
                for idx, callback_event in enumerate(callback_events):
                    if (
                        isinstance(callback_event, dict)
                        and "__root__" in callback_event
                    ):
                        callback_events[idx] = callback_event["__root__"]
    else:
        raise AssertionError(
            f"Expected a InvokeBaseResponse or BatchBaseResponse got: {type(model)}"
        )

    return JSONResponse(content=obj)


def _add_callbacks(
    config: RunnableConfig, callbacks: Sequence[AsyncCallbackHandler]
) -> None:
    """Add the callback aggregator to the config."""
    if "callbacks" not in config:
        config["callbacks"] = []
    config["callbacks"].extend(callbacks)


class _APIHandler:
    """Implementation of the various API endpoints for a runnable server.

    This is a private class whose API is expected to change.

    Currently, the sole role of the handler at the moment is to separate the
    implementation of the endpoints from the logic that registers them on
    a FastAPI app and logic that adds them to the FastAPI OpenAPI docs.
    """

    def __init__(
        self,
        runnable: Runnable,
        *,
        path: str = "",
        prefix: str = "",
        input_type: Union[Type, Literal["auto"], BaseModel] = "auto",
        output_type: Union[Type, Literal["auto"], BaseModel] = "auto",
        config_keys: Sequence[str] = ("configurable",),
        include_callback_events: bool = False,
        enable_feedback_endpoint: bool = False,
        per_req_config_modifier: Optional[PerRequestConfigModifier] = None,
        stream_log_name_allow_list: Optional[Sequence[str]] = None,
    ) -> None:
        """Create an API handler for the given runnable.

        The API Handler has implementations for the various endpoints
        associated with the runnable. The endpoints use Request and Response
        objects from Starlette (same as fast API). The API handler
        does complete validation of the request objects.

        Response validation can be added by the user.

        Args:
            runnable: The runnable to serve.
            path: The path to serve the runnable under.
            prefix: Any prefix that may need to be added to the path
                to get the full URL for the runnable.
                This is relevant when the runnable is added to an APIRouter.
            input_type: type to use for input validation.
                Default is "auto" which will use the InputType of the runnable.
                User is free to provide a custom type annotation.
                Favor using runnable.with_types(input_type=..., output_type=...)
                instead. This parameter may get deprecated!
            output_type: type to use for output validation.
                Default is "auto" which will use the OutputType of the runnable.
                User is free to provide a custom type annotation.
                Favor using runnable.with_types(input_type=..., output_type=...)
                instead. This parameter may get deprecated!
            config_keys: list of config keys that will be accepted, by default
                will accept `configurable` key in the config. Will only be used
                if the runnable is configurable. Cannot configure run_name,
                which is set by default to the path of the API.
            include_callback_events: Whether to include callback events in the response.
                If true, the client will be able to show trace information
                including events that occurred on the server side.
                Be sure not to include any sensitive information in the callback events.
            enable_feedback_endpoint: Whether to enable an endpoint for logging feedback
                to LangSmith. Disabled by default. If this flag is disabled or LangSmith
                tracing is not enabled for the runnable, then 4xx errors will be thrown
                when accessing the feedback endpoint
            per_req_config_modifier: optional function that can be used to update the
                RunnableConfig for a given run based on the raw request. This is useful,
                for example, if the user wants to pass in a header containing
                credentials to a runnable. The RunnableConfig is presented in its
                dictionary form. Note that only keys in `config_keys` will be
                modifiable by this function.
        """
        if importlib.util.find_spec("sse_starlette") is None:
            raise ImportError(
                "sse_starlette must be installed to implement the stream and "
                "stream_log endpoints. "
                "Use `pip install sse_starlette` to install."
            )

        if "run_name" in config_keys:
            raise ValueError(
                "Cannot configure run_name. Please remove it from config_keys."
            )

        self._config_keys = config_keys
        self._path = path
        self._base_url = prefix + path
        self._include_callback_events = include_callback_events
        self._per_req_config_modifier = per_req_config_modifier
        self._serializer = WellKnownLCSerializer()
        self._enable_feedback_endpoint = enable_feedback_endpoint
        self._stream_log_name_allow_list = stream_log_name_allow_list

        # Client is patched using mock.patch, if changing the names
        # remember to make relevant updates in the unit tests.
        self._langsmith_client = (
            ls_client.Client()
            if tracing_is_enabled() and enable_feedback_endpoint
            else None
        )

        with_types = {}

        if input_type != "auto":
            with_types["input_type"] = input_type
        if output_type != "auto":
            with_types["output_type"] = output_type

        if with_types:
            runnable = runnable.with_types(**with_types)

        self._runnable = runnable

        model_namespace = _replace_non_alphanumeric_with_underscores(path.strip("/"))

        input_type_ = _resolve_model(
            runnable.get_input_schema(), "Input", model_namespace
        )

        output_type_ = _resolve_model(
            runnable.get_output_schema(),
            "Output",
            model_namespace,
        )

        self._ConfigPayload = _add_namespace_to_model(
            model_namespace, runnable.config_schema(include=config_keys)
        )

        self._InvokeRequest = create_invoke_request_model(
            model_namespace, input_type_, self._ConfigPayload
        )

        self._BatchRequest = create_batch_request_model(
            model_namespace, input_type_, self._ConfigPayload
        )
        self._StreamRequest = create_stream_request_model(
            model_namespace, input_type_, self._ConfigPayload
        )
        self._StreamLogRequest = create_stream_log_request_model(
            model_namespace, input_type_, self._ConfigPayload
        )
        # Generate the response models
        self._InvokeResponse = create_invoke_response_model(
            model_namespace, output_type_
        )
        self._BatchResponse = create_batch_response_model(model_namespace, output_type_)

        def _route_name(name: str) -> str:
            """Return the route name with the given name."""
            return f"{path.strip('/')} {name}" if path else name

        self._route_name = _route_name

        def _route_name_with_config(name: str) -> str:
            """Return the route name with the given name."""
            return (
                f"{path.strip('/')} {name} with config"
                if path
                else f"{name} with config"
            )

        self._route_name_with_config = _route_name_with_config

    @property
    def InvokeRequest(self) -> Type[BaseModel]:
        """Return the invoke request model."""
        return self._InvokeRequest

    @property
    def BatchRequest(self) -> Type[BaseModel]:
        """Return the batch request model."""
        return self._BatchRequest

    @property
    def StreamRequest(self) -> Type[BaseModel]:
        """Return the stream request model."""
        return self._StreamRequest

    @property
    def StreamLogRequest(self) -> Type[BaseModel]:
        """Return the stream log request model."""
        return self._StreamLogRequest

    @property
    def InvokeResponse(self) -> Type[BaseModel]:
        """Return the invoke response model."""
        return self._InvokeResponse

    @property
    def BatchResponse(self) -> Type[BaseModel]:
        """Return the batch response model."""
        return self._BatchResponse

    async def _get_config_and_input(
        self, request: Request, config_hash: str, *, endpoint: Optional[str] = None
    ) -> Tuple[RunnableConfig, Any]:
        """Extract the config and input from the request, validating the request."""
        try:
            body = await request.json()
        except json.JSONDecodeError:
            raise RequestValidationError(errors=["Invalid JSON body"])
        try:
            body = InvokeRequestShallowValidator.validate(body)

            # Merge the config from the path with the config from the body.
            user_provided_config = _unpack_request_config(
                config_hash,
                body.config,
                config_keys=self._config_keys,
                model=self._ConfigPayload,
                request=request,
                per_req_config_modifier=self._per_req_config_modifier,
            )
            config = _update_config_with_defaults(
                self._path, user_provided_config, request, endpoint=endpoint
            )
            # Unpack the input dynamically using the input schema of the runnable.
            # This takes into account changes in the input type when
            # using configuration.
            schema = self._runnable.with_config(config).input_schema
            input_ = schema.validate(body.input)
            return config, _unpack_input(input_)
        except ValidationError as e:
            raise RequestValidationError(e.errors(), body=body)

    async def invoke(
        self,
        request: Request,
        config_hash: str = "",
    ) -> Response:
        """Invoke the runnable with the given input and config."""
        # We do not use the InvokeRequest model here since configurable runnables
        # have dynamic schema -- so the validation below is a bit more involved.
        config, input_ = await self._get_config_and_input(
            request, config_hash, endpoint="invoke"
        )

        event_aggregator = AsyncEventAggregatorCallback()
        _add_callbacks(config, [event_aggregator])
        output = await self._runnable.ainvoke(input_, config=config)

        if self._include_callback_events:
            callback_events = [
                _scrub_exceptions_in_event(event)
                for event in event_aggregator.callback_events
            ]
        else:
            callback_events = []

        return _json_encode_response(
            self._InvokeResponse(
                output=self._serializer.dumpd(output),
                # Callbacks are scrubbed and exceptions are converted to
                # serializable format before returned in the response.
                callback_events=callback_events,
                metadata=SingletonResponseMetadata(
                    run_id=_get_base_run_id_as_str(event_aggregator)
                ),
            ),
        )

    async def batch(
        self,
        request: Request,
        config_hash: str = "",
    ) -> Response:
        """Invoke the runnable with the given inputs and config."""
        try:
            body = await request.json()
        except json.JSONDecodeError:
            raise RequestValidationError(errors=["Invalid JSON body"])

        with _with_validation_error_translation():
            body = BatchRequestShallowValidator.validate(body)
            config = body.config

            # First unpack the config
            if isinstance(config, list):
                if len(config) != len(body.inputs):
                    raise HTTPException(
                        422,
                        f"Number of configs ({len(config)}) must "
                        f"match number of inputs ({len(body.inputs)})",
                    )

                configs = [
                    _unpack_request_config(
                        config_hash,
                        config,
                        config_keys=self._config_keys,
                        model=self._ConfigPayload,
                        request=request,
                        per_req_config_modifier=self._per_req_config_modifier,
                    )
                    for config in config
                ]
            elif isinstance(config, dict):
                configs = _unpack_request_config(
                    config_hash,
                    config,
                    config_keys=self._config_keys,
                    model=self._ConfigPayload,
                    request=request,
                    per_req_config_modifier=self._per_req_config_modifier,
                )
            else:
                raise HTTPException(
                    422, "Value for 'config' key must be a dict or list if provided"
                )

        inputs_ = body.inputs

        # Make sure that the number of configs matches the number of inputs
        # Since we'll be adding callbacks to the configs.

        configs_ = [
            {k: v for k, v in config_.items() if k in self._config_keys}
            for config_ in get_config_list(configs, len(inputs_))
        ]

        inputs = [
            _unpack_input(
                self._runnable.with_config(config_).input_schema.validate(input_)
            )
            for config_, input_ in zip(configs_, inputs_)
        ]

        # Update the configuration with callbacks
        aggregators = [AsyncEventAggregatorCallback() for _ in range(len(inputs))]

        final_configs = []
        for config_, aggregator in zip(configs_, aggregators):
            _add_callbacks(config_, [aggregator])
            final_configs.append(
                _update_config_with_defaults(
                    self._path, config_, request, endpoint="batch"
                )
            )

        output = await self._runnable.abatch(inputs, config=final_configs)

        if self._include_callback_events:
            callback_events = [
                # Scrub sensitive information and convert
                # exceptions to serializable format
                [
                    _scrub_exceptions_in_event(event)
                    for event in aggregator.callback_events
                ]
                for aggregator in aggregators
            ]
        else:
            callback_events = []

        obj = self._BatchResponse(
            output=self._serializer.dumpd(output),
            callback_events=callback_events,
            metadata=BatchResponseMetadata(
                run_ids=[_get_base_run_id_as_str(agg) for agg in aggregators]
            ),
        )
        return _json_encode_response(obj)

    async def stream(
        self,
        request: Request,
        config_hash: str = "",
    ) -> EventSourceResponse:
        """Invoke the runnable stream the output.

        See documentation for endpoint at the end of the file.
        It's attached to _stream_docs endpoint.
        """
        err_event = {}
        validation_exception: Optional[BaseException] = None
        try:
            config, input_ = await self._get_config_and_input(
                request, config_hash, endpoint="stream"
            )
        except BaseException as e:
            validation_exception = e
            if isinstance(e, RequestValidationError):
                err_event = {
                    "event": "error",
                    "data": json.dumps(
                        {"status_code": 422, "message": repr(e.errors())}
                    ),
                }
            else:
                err_event = {
                    "event": "error",
                    # Do not expose the error message to the client since
                    # the message may contain sensitive information.
                    "data": json.dumps(
                        {"status_code": 500, "message": "Internal Server Error"}
                    ),
                }

        async def _stream() -> AsyncIterator[dict]:
            """Stream the output of the runnable."""
            if validation_exception:
                yield err_event
                if isinstance(validation_exception, RequestValidationError):
                    return
                else:
                    raise AssertionError(
                        "Internal server error"
                    ) from validation_exception

            try:
                config_w_callbacks = config.copy()
                event_aggregator = AsyncEventAggregatorCallback()
                _add_callbacks(config_w_callbacks, [event_aggregator])
                has_sent_metadata = False
                async for chunk in self._runnable.astream(
                    input_,
                    config=config_w_callbacks,
                ):
                    if not has_sent_metadata and event_aggregator.callback_events:
                        yield {
                            "event": "metadata",
                            "data": json.dumps(
                                {
                                    "run_id": _get_base_run_id_as_str(event_aggregator),
                                }
                            ),
                        }
                        has_sent_metadata = True

                    yield {
                        # EventSourceResponse expects a string for data
                        # so after serializing into bytes, we decode into utf-8
                        # to get a string.
                        "data": self._serializer.dumps(chunk).decode("utf-8"),
                        "event": "data",
                    }
                yield {"event": "end"}
            except BaseException:
                yield {
                    "event": "error",
                    # Do not expose the error message to the client since
                    # the message may contain sensitive information.
                    # We'll add client side errors for validation as well.
                    "data": json.dumps(
                        {"status_code": 500, "message": "Internal Server Error"}
                    ),
                }
                raise

        return EventSourceResponse(_stream())

    async def stream_log(
        self,
        request: Request,
        config_hash: str = "",
    ) -> EventSourceResponse:
        """Invoke the runnable stream_log the output.

        View documentation for endpoint at the end of the file.
        It's attached to _stream_log_docs endpoint.
        """
        err_event = {}
        validation_exception: Optional[BaseException] = None
        try:
            config, input_ = await self._get_config_and_input(
                request, config_hash, endpoint="stream_log"
            )
        except BaseException as e:
            validation_exception = e
            if isinstance(e, RequestValidationError):
                err_event = {
                    "event": "error",
                    "data": json.dumps(
                        {"status_code": 422, "message": repr(e.errors())}
                    ),
                }
            else:
                err_event = {
                    "event": "error",
                    # Do not expose the error message to the client since
                    # the message may contain sensitive information.
                    "data": json.dumps(
                        {"status_code": 500, "message": "Internal Server Error"}
                    ),
                }

        try:
            body = await request.json()
            with _with_validation_error_translation():
                stream_log_request = StreamLogParameters(**body)
        except json.JSONDecodeError:
            # Body as text
            validation_exception = RequestValidationError(errors=["Invalid JSON body"])
            err_event = {
                "event": "error",
                "data": json.dumps(
                    {"status_code": 422, "message": "Invalid JSON body"}
                ),
            }
        except RequestValidationError as e:
            validation_exception = e
            err_event = {
                "event": "error",
                "data": json.dumps({"status_code": 422, "message": repr(e.errors())}),
            }

        async def _stream_log() -> AsyncIterator[dict]:
            """Stream the output of the runnable."""
            if validation_exception:
                yield err_event
                if isinstance(validation_exception, RequestValidationError):
                    return
                else:
                    raise AssertionError(
                        "Internal server error"
                    ) from validation_exception

            try:
                async for chunk in self._runnable.astream_log(
                    input_,
                    config=config,
                    diff=True,
                    include_names=stream_log_request.include_names,
                    include_types=stream_log_request.include_types,
                    include_tags=stream_log_request.include_tags,
                    exclude_names=stream_log_request.exclude_names,
                    exclude_types=stream_log_request.exclude_types,
                    exclude_tags=stream_log_request.exclude_tags,
                ):
                    if not isinstance(chunk, RunLogPatch):
                        raise AssertionError(
                            f"Expected a RunLog instance got {type(chunk)}"
                        )
                    if (
                        self._stream_log_name_allow_list is None
                        or self._runnable.config.get("run_name")
                        in self._stream_log_name_allow_list
                    ):
                        data = {
                            "ops": chunk.ops,
                        }

                        # Temporary adapter
                        yield {
                            # EventSourceResponse expects a string for data
                            # so after serializing into bytes, we decode into utf-8
                            # to get a string.
                            "data": self._serializer.dumps(data).decode("utf-8"),
                            "event": "data",
                        }
                yield {"event": "end"}
            except BaseException:
                yield {
                    "event": "error",
                    # Do not expose the error message to the client since
                    # the message may contain sensitive information.
                    # We'll add client side errors for validation as well.
                    "data": json.dumps(
                        {"status_code": 500, "message": "Internal Server Error"}
                    ),
                }
                raise

        return EventSourceResponse(_stream_log())

    async def input_schema(self, request: Request, config_hash: str = "") -> Any:
        """Return the input schema of the runnable."""
        with _with_validation_error_translation():
            user_provided_config = _unpack_request_config(
                config_hash,
                config_keys=self._config_keys,
                model=self._ConfigPayload,
                request=request,
                per_req_config_modifier=self._per_req_config_modifier,
            )
            config = _update_config_with_defaults(
                self._path, user_provided_config, request
            )

        return self._runnable.get_input_schema(config).schema()

    async def output_schema(self, request: Request, config_hash: str = "") -> Any:
        """Return the output schema of the runnable."""
        with _with_validation_error_translation():
            user_provided_config = _unpack_request_config(
                config_hash,
                config_keys=self._config_keys,
                model=self._ConfigPayload,
                request=request,
                per_req_config_modifier=self._per_req_config_modifier,
            )
            config = _update_config_with_defaults(
                self._path, user_provided_config, request
            )
        return self._runnable.get_output_schema(config).schema()

    async def config_schema(self, request: Request, config_hash: str = "") -> Any:
        """Return the config schema of the runnable."""
        with _with_validation_error_translation():
            user_provided_config = _unpack_request_config(
                config_hash,
                config_keys=self._config_keys,
                model=self._ConfigPayload,
                request=request,
                per_req_config_modifier=self._per_req_config_modifier,
            )
            config = _update_config_with_defaults(
                self._path, user_provided_config, request
            )
        return (
            self._runnable.with_config(config)
            .config_schema(include=self._config_keys)
            .schema()
        )

    async def playground(
        self, file_path: str, request: Request, config_hash: str = ""
    ) -> Any:
        """Return the playground of the runnable."""
        with _with_validation_error_translation():
            user_provided_config = _unpack_request_config(
                config_hash,
                config_keys=self._config_keys,
                model=self._ConfigPayload,
                request=request,
                per_req_config_modifier=self._per_req_config_modifier,
            )

            config = _update_config_with_defaults(
                self._path, user_provided_config, request
            )

        feedback_enabled = tracing_is_enabled() and self._enable_feedback_endpoint

        if self._base_url.endswith("/"):
            playground_url = self._base_url + "playground"
        else:
            playground_url = self._base_url + "/playground"

        return await serve_playground(
            self._runnable.with_config(config),
            self._runnable.with_config(config).input_schema,
            self._config_keys,
            playground_url,
            file_path,
            feedback_enabled,
        )

    async def create_feedback(
        self, feedback_create_req: FeedbackCreateRequest, config_hash: str = ""
    ) -> Feedback:
        """Send feedback on an individual run to langsmith

        Note that a successful response means that feedback was successfully
        submitted. It does not guarantee that the feedback is recorded by
        langsmith. Requests may be silently rejected if they are
        unauthenticated or invalid by the server.
        """

        if not tracing_is_enabled() or not self._enable_feedback_endpoint:
            raise HTTPException(
                400,
                "The feedback endpoint is only accessible when LangSmith is "
                + "enabled on your LangServe server.",
            )

        feedback_from_langsmith = self._langsmith_client.create_feedback(
            feedback_create_req.run_id,
            feedback_create_req.key,
            score=feedback_create_req.score,
            value=feedback_create_req.value,
            comment=feedback_create_req.comment,
            source_info={
                "from_langserve": True,
            },
        )

        # We purposefully select out fields from langsmith so that we don't
        # fail validation if langsmith adds extra fields. We prefer this over
        # using "Extra.allow" in pydantic since syntax changes between pydantic
        # 1.x and 2.x for this functionality
        return Feedback(
            id=str(feedback_from_langsmith.id),
            run_id=str(feedback_from_langsmith.run_id),
            created_at=str(feedback_from_langsmith.created_at),
            modified_at=str(feedback_from_langsmith.modified_at),
            key=str(feedback_from_langsmith.key),
            score=feedback_from_langsmith.score,
            value=feedback_from_langsmith.value,
            comment=feedback_from_langsmith.comment,
        )

    async def _check_feedback_enabled(self, config_hash: str = "") -> None:
        """Check if feedback is enabled for the runnable.

        This endpoint is private since it will be deprecated in the future.

        """
        if not (await self.check_feedback_enabled(config_hash)):
            raise HTTPException(
                400,
                "The feedback endpoint is only accessible when LangSmith is "
                + "enabled on your LangServe server.",
            )

    async def check_feedback_enabled(self, config_hash: str = "") -> bool:
        """Check if feedback is enabled for the runnable."""
        return self._enable_feedback_endpoint or not tracing_is_enabled()


_MODEL_REGISTRY = {}
_SEEN_NAMES = set()



from __future__ import annotations

import uuid
from typing import Any, Dict, List, Optional, Sequence
from uuid import UUID

from langchain.callbacks.base import AsyncCallbackHandler
from langchain.callbacks.manager import (
    BaseRunManager,
    ahandle_event,
    handle_event,
)
from langchain.schema import AgentAction, AgentFinish, BaseMessage, Document, LLMResult
from typing_extensions import TypedDict


class CallbackEventDict(TypedDict, total=False):
    """A dictionary representation of a callback event."""

    type: str
    parent_run_id: Optional[UUID]
    run_id: UUID


class AsyncEventAggregatorCallback(AsyncCallbackHandler):
    """A callback handler that aggregates all the events that have been called.

    This callback handler aggregates all the events that have been called placing
    them in a single mutable list.

    This callback handler is not threading safe, and is meant to be used in an async
    context only.
    """

    def __init__(self) -> None:
        """Get a list of all the callback events that have been called."""
        super().__init__()
        # Callback events is a mutable state that is used only in an async context,
        # so it should be safe to mutate without the usage of a lock.
        self.callback_events: List[CallbackEventDict] = []

    def log_callback(self, event: CallbackEventDict) -> None:
        """Log the callback event."""
        self.callback_events.append(event)

    async def on_chat_model_start(
        self,
        serialized: Dict[str, Any],
        messages: List[List[BaseMessage]],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        """Attempt to serialize the callback event."""
        self.log_callback(
            {
                "type": "on_chat_model_start",
                "serialized": serialized,
                "messages": messages,
                "run_id": run_id,
                "parent_run_id": parent_run_id,
                "tags": tags,
                "metadata": metadata,
                "kwargs": kwargs,
            }
        )

    async def on_chain_start(
        self,
        serialized: Dict[str, Any],
        inputs: Dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        """Attempt to serialize the callback event."""
        self.log_callback(
            {
                "type": "on_chain_start",
                "serialized": serialized,
                "inputs": inputs,
                "run_id": run_id,
                "parent_run_id": parent_run_id,
                "tags": tags,
                "metadata": metadata,
                "kwargs": kwargs,
            }
        )

    async def on_chain_end(
        self,
        outputs: Any,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        self.log_callback(
            {
                "type": "on_chain_end",
                "outputs": outputs,
                "run_id": run_id,
                "parent_run_id": parent_run_id,
                "tags": tags,
                "kwargs": kwargs,
            }
        )

    async def on_chain_error(
        self,
        error: BaseException,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        self.log_callback(
            {
                "type": "on_chain_error",
                "error": error,
                "run_id": run_id,
                "parent_run_id": parent_run_id,
                "tags": tags,
                "kwargs": kwargs,
            }
        )

    async def on_retriever_start(
        self,
        serialized: Dict[str, Any],
        query: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        self.log_callback(
            {
                "type": "on_retriever_start",
                "serialized": serialized,
                "query": query,
                "run_id": run_id,
                "parent_run_id": parent_run_id,
                "tags": tags,
                "metadata": metadata,
                "kwargs": kwargs,
            }
        )

    async def on_retriever_end(
        self,
        documents: Sequence[Document],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        self.log_callback(
            {
                "type": "on_retriever_end",
                "documents": documents,
                "run_id": run_id,
                "parent_run_id": parent_run_id,
                "tags": tags,
                "kwargs": kwargs,
            }
        )

    async def on_retriever_error(
        self,
        error: BaseException,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        self.log_callback(
            {
                "type": "on_retriever_error",
                "error": error,
                "run_id": run_id,
                "parent_run_id": parent_run_id,
                "tags": tags,
                "kwargs": kwargs,
            }
        )

    async def on_tool_start(
        self,
        serialized: Dict[str, Any],
        input_str: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        self.log_callback(
            {
                "type": "on_tool_start",
                "serialized": serialized,
                "input_str": input_str,
                "run_id": run_id,
                "parent_run_id": parent_run_id,
                "tags": tags,
                "metadata": metadata,
                "kwargs": kwargs,
            }
        )

    async def on_tool_end(
        self,
        output: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        self.log_callback(
            {
                "type": "on_tool_end",
                "output": output,
                "run_id": run_id,
                "parent_run_id": parent_run_id,
                "tags": tags,
                "kwargs": kwargs,
            }
        )

    async def on_tool_error(
        self,
        error: BaseException,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        self.log_callback(
            {
                "type": "on_tool_error",
                "error": error,
                "run_id": run_id,
                "parent_run_id": parent_run_id,
                "tags": tags,
                "kwargs": kwargs,
            }
        )

    async def on_agent_action(
        self,
        action: AgentAction,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        self.log_callback(
            {
                "type": "on_agent_action",
                "action": action,
                "run_id": run_id,
                "parent_run_id": parent_run_id,
                "tags": tags,
                "kwargs": kwargs,
            }
        )

    async def on_agent_finish(
        self,
        finish: AgentFinish,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        self.log_callback(
            {
                "type": "on_agent_finish",
                "finish": finish,
                "run_id": run_id,
                "parent_run_id": parent_run_id,
                "tags": tags,
                "kwargs": kwargs,
            }
        )

    async def on_llm_start(
        self,
        serialized: Dict[str, Any],
        prompts: List[str],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        self.log_callback(
            {
                "type": "on_llm_start",
                "serialized": serialized,
                "prompts": prompts,
                "run_id": run_id,
                "parent_run_id": parent_run_id,
                "tags": tags,
                "metadata": metadata,
                "kwargs": kwargs,
            }
        )

    async def on_llm_end(
        self,
        response: LLMResult,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        self.log_callback(
            {
                "type": "on_llm_end",
                "response": response,
                "run_id": run_id,
                "parent_run_id": parent_run_id,
                "tags": tags,
                "kwargs": kwargs,
            }
        )

    async def on_llm_error(
        self,
        error: BaseException,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        self.log_callback(
            {
                "type": "on_llm_error",
                "error": error,
                "run_id": run_id,
                "parent_run_id": parent_run_id,
                "tags": tags,
                "kwargs": kwargs,
            }
        )


def replace_uuids(
    callback_events: Sequence[CallbackEventDict],
) -> List[CallbackEventDict]:
    """Replace uids in the event callbacks with new uids.

    This function mutates the event callback events in place.

    Args:
        callback_events: A list of event callbacks.
    """
    # Create a dictionary to store mappings from old UID to new UID
    uid_mapping: dict = {}

    updated_events = []

    # Iterate through the list of event callbacks
    for event in callback_events:
        updated_event = event.copy()
        # Replace UIDs in the 'run_id' field
        if "run_id" in updated_event and updated_event["run_id"] is not None:
            if updated_event["run_id"] not in uid_mapping:
                # Generate a new UUID
                new_uid = uuid.uuid4()
                uid_mapping[updated_event["run_id"]] = new_uid
            # Replace the old UID with the new one
            updated_event["run_id"] = uid_mapping[updated_event["run_id"]]

        # Replace UIDs in the 'parent_run_id' field if it's not None
        if (
            "parent_run_id" in updated_event
            and updated_event["parent_run_id"] is not None
        ):
            if updated_event["parent_run_id"] not in uid_mapping:
                # Generate a new UUID
                new_uid = uuid.uuid4()
                uid_mapping[updated_event["parent_run_id"]] = new_uid
            # Replace the old UID with the new one
            updated_event["parent_run_id"] = uid_mapping[updated_event["parent_run_id"]]
        updated_events.append(updated_event)
    return updated_events


# Mapping from event name to ignore condition name
NAME_TO_IGNORE_CONDITION = {
    "on_retry": "ignore_retry",
    "on_text": None,
    "on_agent_action": "ignore_agent",
    "on_agent_finish": "ignore_agent",
    "on_llm_start": "ignore_llm",
    "on_llm_end": "ignore_llm",
    "on_llm_error": "ignore_llm",
    "on_chain_start": "ignore_chain",
    "on_chain_end": "ignore_chain",
    "on_chain_error": "ignore_chain",
    "on_chat_model_start": "ignore_chat_model",
    "on_tool_start": "ignore_agent",
    "on_tool_end": "ignore_agent",
    "on_tool_error": "ignore_agent",
    "on_retriever_start": "ignore_retriever",
    "on_retriever_end": "ignore_retriever",
    "on_retriever_error": "ignore_retriever",
}


async def ahandle_callbacks(
    callback_manager: BaseRunManager,
    callback_events: Sequence[CallbackEventDict],
) -> None:
    """Invoke all the callback handlers with the given callback events."""
    callback_events = replace_uuids(callback_events)

    # 1. Do I need inheritable handlers
    for event in callback_events:
        if event["parent_run_id"] is None:  # How do we make sure it's None!?
            event["parent_run_id"] = callback_manager.run_id

        event_data = {key: value for key, value in event.items() if key != "type"}

        await ahandle_event(
            # Unpacking like this may not work
            callback_manager.handlers,
            event["type"],
            ignore_condition_name=NAME_TO_IGNORE_CONDITION.get(event["type"], None),
            **event_data,
        )


def handle_callbacks(
    callback_manager: BaseRunManager,
    callback_events: Sequence[CallbackEventDict],
) -> None:
    """Invoke all the callback handlers with the given callback events."""
    callback_events = replace_uuids(callback_events)

    for event in callback_events:
        if event["parent_run_id"] is None:  # How do we make sure it's None!?
            event["parent_run_id"] = callback_manager.run_id

        event_data = {key: value for key, value in event.items() if key != "type"}

        handle_event(
            # Unpacking like this may not work
            callback_manager.handlers,
            event["type"],
            ignore_condition_name=NAME_TO_IGNORE_CONDITION.get(event["type"], None),
            **event_data,
        )
